/******************************************************************************
  Copyright (c) 1994, 1995, 1996 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* Modified by Christopher Unkel for WinMOO: */
/**********************************************************************
 Copyright (c) 1996, 1997 Christopher Unkel.  All rights reserved.
 
 Use and copying of the software and preparation and distribution of
 derivative works is permitted provided this copyright notice is
 left intact.  Any distribution of this software or derivative works
 must comply with all applicable United States export control laws.
 
 This software is provided AS IS, without express or implied warranty
 of any kind.  Christopher Unkel makes no representations concerning
 either the merchantability of this software or the suitability of
 this software of any particular purpose.
 **********************************************************************/

/* This module was more or less entirely rewritten for WinMOO.  It uses a second
   thread, rather than two extra processes, to perform name-lookups with
   timeouts.  It also caches lookup results. */

#include "options.h"

#include <windows.h>
#include <process.h>
#include <stddef.h>

#include "my-stdio.h"
#include "my-time.h"

#include "my-signal.h"
#include "my-stdlib.h"
#include "my-unistd.h"
#include "my-inet.h"		/* inet_addr() */
#include "my-in.h"		/* struct sockaddr_in, INADDR_ANY, htons(),
				 * htonl(), ntohl(), struct in_addr */
#ifndef _WINSOCKAPI_
#include <winsock.h>		/* struct hostent, gethostbyaddr() */
#endif

#include "my-socket.h"		/* AF_INET */
#include "my-wait.h"
#include "my-string.h"
#include <errno.h>

#include "config.h"
#include "log.h"
#include "server.h"
#include "storage.h"
#include "timers.h"


#define MAX_CACHE_LENGTH 250
#define CACHE_PRUNE_LENGTH 25
#define MAX_ENTRY_AGE 86400
#define MAX_FAILED_LOOKUP_AGE 300

typedef struct name_entry_s {
  char *hostname;
  unsigned hostaddr;
  time_t lookup_time;
  int seen;
  struct name_entry_s *next;
} name_entry;

static name_entry *host_to_ip, *ip_to_host;

static HANDLE host_to_ip_mutex, ip_to_host_mutex;

static int num_host_to_ip=0, num_ip_to_host=0;

static void lookup_addr_from_name_proc(void * data);
char *my_inet_ntoa (unsigned a);
void lookup_name_from_addr_proc(void * data);


static void
lookup_addr_from_name_proc(void * data)
{
  char *domain_name=data;
  struct hostent *hostdata;
  name_entry *foo, **bar;

  hostdata=gethostbyname(domain_name);

  WaitForSingleObject(host_to_ip_mutex, INFINITE);

  /* eliminate any duplicates */
  bar=&host_to_ip;
  foo=host_to_ip;
  while (foo) {
    if (!strcmp(foo->hostname, domain_name)) {
      *bar=foo->next;
      free(foo->hostname);
      free(foo);
      foo=*bar;
      num_host_to_ip--;
    } else {
      bar=&foo->next;
      foo=foo->next;
    }
  }

  foo=(name_entry *)malloc(sizeof(name_entry));
  foo->hostname=domain_name;
  if (hostdata)
    foo->hostaddr=*(unsigned *) hostdata->h_addr_list[0];
  else
    foo->hostaddr=0;
  foo->lookup_time=time(NULL);
  foo->seen=0;
  foo->next=host_to_ip;
  host_to_ip=foo;

  num_host_to_ip++;

  ReleaseMutex(host_to_ip_mutex);
}


unsigned
lookup_addr_from_name(const char *hostname, int timeout)
{
  static unsigned host_addr;
  char *tempchar;
  unsigned long thread;
  int i;
  int lookforold=0;

  name_entry *foo, *foobar, **bar;

  host_addr=inet_addr(hostname);
  if (host_addr!=INADDR_NONE)
    return host_addr;

  WaitForSingleObject(host_to_ip_mutex, INFINITE);
  foo=host_to_ip;
  bar=&host_to_ip;
  while (foo) {
    foo->seen=1;
    if (!strcmp(foo->hostname, hostname)) {
      /* move to front of list */
      *bar=foo->next;
      foo->next=host_to_ip;
      host_to_ip=foo;

      if (time(0)-(foo->lookup_time) > MAX_ENTRY_AGE) {
        lookforold=1;
        break;
      }

      host_addr=foo->hostaddr;
      if (host_addr) {
        ReleaseMutex(host_to_ip_mutex);
        return host_addr;
      }
      else {
        if (time(0)-(foo->lookup_time) > MAX_FAILED_LOOKUP_AGE)
          break;

        ReleaseMutex(host_to_ip_mutex);
        return 0;
      }
    } else {
      bar=&foo->next;
      foo=foo->next;
    }
  }

  /* cleanup here */
  if (num_host_to_ip > MAX_CACHE_LENGTH) {
    foo=host_to_ip;
    i=1;
    while (i < num_host_to_ip-CACHE_PRUNE_LENGTH) {
      foo=foo->next;
      i++;
    }

    num_host_to_ip=num_host_to_ip-CACHE_PRUNE_LENGTH;

    foobar=foo->next;
    foo->next=NULL;
    foo=foobar;
    
    while (foo) {
      foobar=foo->next;
      free(foo->hostname);
      free(foo);
      foo=foobar;
    }
  }
  
  ReleaseMutex(host_to_ip_mutex);

  tempchar=malloc(strlen(hostname)+1);
  strcpy(tempchar, hostname);

  thread=_beginthread(lookup_addr_from_name_proc, 0, tempchar);

  if (thread==-1)
    return 0;

  WaitForSingleObject((HANDLE)thread, timeout*1000);

  WaitForSingleObject(host_to_ip_mutex, INFINITE);

  foo=host_to_ip;
  while (foo && (lookforold || !foo->seen)) {
    foo->seen=1;
    if (!strcmp(foo->hostname, hostname)) {
      host_addr=foo->hostaddr;
      ReleaseMutex(host_to_ip_mutex);

      if (host_addr)
        return host_addr;
      else
        return 0;
    } else
      foo=foo->next;
  }

  ReleaseMutex(host_to_ip_mutex);
  return 0;
}

char *
my_inet_ntoa (unsigned a)
{
  static char buffer[30];
  char *tc;
	a = ntohl(a);

	sprintf(buffer, "%u.%u.%u.%u",
		(unsigned) (a >> 24) & 0xff, (unsigned) (a >> 16) & 0xff,
		(unsigned) (a >> 8) & 0xff, (unsigned) a & 0xff);

  tc = (char *)malloc(strlen(buffer)+1);
  strcpy(tc, buffer);
  return tc;
}

void
lookup_name_from_addr_proc(void * data)
{
  unsigned hostaddr=*(unsigned *)data;
  struct hostent *hostdata;
  name_entry *foo, **bar;


  free(data);
  hostdata=gethostbyaddr((char *)&hostaddr, 4, PF_INET);

  WaitForSingleObject(ip_to_host_mutex, INFINITE);

  /* eliminate any duplicates */
  bar=&ip_to_host;
  foo=ip_to_host;
  while (foo) {
    if (foo->hostaddr==hostaddr) {
      *bar=foo->next;
      free(foo->hostname);
      free(foo);
      foo=*bar;
      num_ip_to_host--;
    } else {
      bar=&foo->next;
      foo=foo->next;
    }
  }

  foo=(name_entry *)malloc(sizeof(name_entry));
  foo->hostaddr=hostaddr;
  if (hostdata) {
    foo->hostname=(char *) malloc(strlen(hostdata->h_name)+1);
    strcpy(foo->hostname, hostdata->h_name);
  } else
    foo->hostname=NULL;
  foo->lookup_time=time(NULL);
  foo->seen=0;
  foo->next=ip_to_host;
  ip_to_host=foo;

  num_ip_to_host++;

  ReleaseMutex(ip_to_host_mutex);
}


const char *
lookup_name_from_addr(struct sockaddr_in *host, unsigned timeout)
{
  unsigned hostaddr=host->sin_addr.s_addr;
  static char *tempchar=0;
  unsigned long thread;
  unsigned long *ti;
  int i;
  int lookforold=0;

  name_entry *foo, *foobar, **bar;

  if (tempchar) {
    free(tempchar);
    tempchar=0;
  }

  WaitForSingleObject(ip_to_host_mutex, INFINITE);
  foo=ip_to_host;
  bar=&ip_to_host;
  while (foo) {
    foo->seen=1;
    if (foo->hostaddr==hostaddr) {
      /* move to front of list */
      *bar=foo->next;
      foo->next=ip_to_host;
      ip_to_host=foo;

      if (time(0)-(foo->lookup_time) > MAX_ENTRY_AGE) {
        lookforold=1;
        break;
      }

      if (foo->hostname) {
        tempchar=(char *)malloc(strlen(foo->hostname)+1);
        strcpy(tempchar,foo->hostname);
        ReleaseMutex(ip_to_host_mutex);
        return tempchar;
      }
      else {
        if (time(0)-(foo->lookup_time) > MAX_FAILED_LOOKUP_AGE)
          break;

        ReleaseMutex(ip_to_host_mutex);
        return tempchar=my_inet_ntoa(hostaddr);
      }
    } else {
      bar=&foo->next;
      foo=foo->next;
    }
  }

  /* cleanup here */
  if (num_ip_to_host > MAX_CACHE_LENGTH) {
    foo=ip_to_host;
    i=1;
    while (i < num_ip_to_host-CACHE_PRUNE_LENGTH) {
      foo=foo->next;
      i++;
    }

    num_ip_to_host=num_ip_to_host-CACHE_PRUNE_LENGTH;

    foobar=foo->next;
    foo->next=NULL;
    foo=foobar;
    
    while (foo) {
      foobar=foo->next;
      free(foo->hostname);
      free(foo);
      foo=foobar;
    }
  }
  
  ReleaseMutex(ip_to_host_mutex);

  ti=(unsigned long*)malloc(sizeof(unsigned long));
  *ti=hostaddr;

  thread=_beginthread(lookup_name_from_addr_proc, 0, ti);

  if (thread==-1)
    return NULL;

  WaitForSingleObject((HANDLE)thread, timeout*1000);

  WaitForSingleObject(ip_to_host_mutex, INFINITE);

  foo=ip_to_host;
  while (foo && (lookforold || !foo->seen)) {
    foo->seen=1;
    if (foo->hostaddr==hostaddr) {
      if (foo->hostname) {
        tempchar=(char *)malloc(strlen(foo->hostname)+1);
        strcpy(tempchar, foo->hostname);
        ReleaseMutex(ip_to_host_mutex);
        return tempchar;
      } else {
        ReleaseMutex(ip_to_host_mutex);
        return tempchar=my_inet_ntoa(hostaddr);
      }
    } else
      foo=foo->next;
  }

  ReleaseMutex(ip_to_host_mutex);
  return tempchar=my_inet_ntoa(hostaddr);
}

int
initialize_name_lookup(void)
{
  host_to_ip=NULL;
  ip_to_host=NULL;

  host_to_ip_mutex=CreateMutex(NULL, 0, NULL);
  ip_to_host_mutex=CreateMutex(NULL, 0, NULL);

  if (!host_to_ip_mutex || !ip_to_host_mutex)
    return 0;

  return 1;
}

char rcsid_name_lookup[] = "$Id: name_lookup.c,v 2.2 1996/02/08 06:59:04 pavel Exp $";

/* $Log: name_lookup.c,v $
 * Revision 2.2  1996/02/08  06:59:04  pavel
 * Renamed err/logf() to errlog/oklog().  Updated copyright notice for 1996.
 * Release 1.8.0beta1.
 *
 * Revision 2.1  1995/12/11  08:11:45  pavel
 * Added missing #include of "my-stdlib.h".  Release 1.8.0alpha2.
 *
 * Revision 2.0  1995/11/30  04:28:09  pavel
 * New baseline version, corresponding to release 1.8.0alpha1.
 *
 * Revision 1.1  1995/11/30  04:27:56  pavel
 * Initial revision
 */
